/* This file is part of the LWRPC library */
#ifndef _LWRPC_INTERNAL_INCLUDED
#define _LWRPC_INTERNAL_INCLUDED

#include <config.h>
#include <pthread.h>
#include <mpi.h>

#ifndef WIN32
#include <unistd.h>

typedef /*unsigned*/ int INT32;
typedef /*unsigned*/ long long INT64;

#else
#include <windows.h>

#endif

typedef unsigned long	VIRT_ADDR;

typedef void (*func_t)();

#define EXIT_PREV	0
#define SAVE_PREV	1

#define MAX_CPUS	16
#define MAX_LWRPC_ARGS	18

#include "uth.h"
#include "utils.h"


typedef struct lwrpc_desc {
	_lock_t					lock;
	uth_t					uth_handler;
	struct lwrpc_desc			*prev;
	struct lwrpc_desc			*next;
	int					vp_id;

	struct lwrpc_desc			*parent;
	int					ndep;

	func_t            			work;
	int					narg;

	int					homenode;
	int					sourcenode;
	int					sourcevpid;
	int					target_queue;

	int					inter_node;
	int					c_routine;
	int					insert_in_front;
	int					insert_private;

	int					rpc_type;
	int					type;
	int					level;

	MPI_Datatype				dtype[MAX_LWRPC_ARGS];
	int					quantity[MAX_LWRPC_ARGS];
	int					callway[MAX_LWRPC_ARGS];
	int					operation[MAX_LWRPC_ARGS];
	INT64					localarg[MAX_LWRPC_ARGS];	/* data (address / value) in the owner node */
	INT64					temparg[MAX_LWRPC_ARGS];	/* data (address / value) in the remote node */
} lwrpc_t;

lwrpc_t *_lwrpc_self();
void _lwrpc_stats();
void _lwrpc_depadd(lwrpc_t *, int ndep);
int _lwrpc_block();
int _lwrpc_depsatisfy (lwrpc_t *);
void _lwrpc_md_init (void);
void _lwrpc_md_end (void);
void _lwrpc_reset_statistics();
void _lwrpc_env_init();
void _lwrpc_yield();
void _lwrpc_opt (int, char **);
void *_lwrpc_worker(void *arg);

#define LWRPC_DEF_STACKPAGES              0
#define LWRPC_DEF_CPUS                    1

void thread_wait();
void thread_signal();

#include "queue.h"
#include "lwrpc_data.h"
#include "lwrpc_mpi_internal.h"





#endif
