/* This file is part of the LWRPC library */
#ifndef __UTILS_H__
#define __UTILS_H__

#define CACHE_LINE_SIZE 128
#define PADS_PER_CACHE_LINE(type)	(CACHE_LINE_SIZE/sizeof(type))

#include <sys/types.h>
#include <errno.h>
#include <pthread.h>

#define POSIX_MUTEX_LOCK        0
#define POSIX_MUTEX_TRYLOCK     1

#define POSIX_LOCK              0

typedef pthread_mutex_t _lock_t;

#define _lock_init(var)			pthread_mutex_init(var, NULL)
#define _lock_try_acquire(var)		pthread_mutex_trylock(var)
#define _lock_release(var)		pthread_mutex_unlock(var)
#define _lock_destroy(var)		pthread_mutex_destroy(var)

#if (POSIX_LOCK == POSIX_MUTEX_LOCK)
#define _lock_acquire(var)		pthread_mutex_lock(var)
#elif (POSIX_LOCK == POSIX_MUTEX_TRYLOCK)
static int pthread_mutex_spinlock (pthread_mutex_t * lock)
{
	while (pthread_mutex_trylock(lock) == EBUSY) { /*noop();*/ }
	return 0;
}
#endif

#define _lock_and_add(address, value) \
{ _lock_acquire(&datalock); *address += value; _lock_release(&datalock); }


#define QUEUE_NODE_FIELDS(nodetype)			\
	nodetype *next, *prev;				


#define QUEUE_DEFINE(nodetype,queue_type) 		\
typedef union _##queue_type {				\
	struct {					\
		_lock_t lock;				\
		nodetype *head;				\
		nodetype *tail;				\
	} q;						\
	char pad[CACHE_LINE_SIZE];			\
} queue_type;


#define _queue_head(_q)		(_q)->q.head
#define _queue_tail(_q)		(_q)->q.tail

#define _queue_init(_q) {				\
	(_q)->q.head = NULL;				\
	(_q)->q.tail = NULL;				\
	_lock_init(&(_q)->q.lock);			\
}


#define _enqueue_head(_q, _e) {				\
	_lock_acquire(&(_q)->q.lock);			\
	(_e)->prev = NULL;				\
	(_e)->next = (_q)->q.head;			\
	if ((_q)->q.head == NULL) {			\
		(_q)->q.head = (_q)->q.tail = (_e);	\
	}						\
	else {						\
		(_q)->q.head = (_e);			\
		(_e)->next->prev = (_e);		\
	}						\
	_lock_release(&(_q)->q.lock);			\
}


#define _enqueue_tail(_q, _e) {				\
	_lock_acquire(&(_q)->q.lock);			\
	(_e)->next = NULL;				\
	(_e)->prev = (_q)->q.tail;			\
	if ((_q)->q.tail == NULL)	{		\
		(_q)->q.tail = (_q)->q.head = (_e);	\
	}						\
	else {						\
		(_q)->q.tail->next = (_e);		\
		(_q)->q.tail = (_e);			\
	}						\
	_lock_release(&(_q)->q.lock);			\
}


#define _dequeue(_q, _thread) {					\
	if ((_q)->q.head != NULL) {				\
		_lock_acquire(&(_q)->q.lock);			\
		(*(_thread)) = (_q)->q.head;			\
		if ((_q)->q.head != NULL) {			\
			(_q)->q.head = (_q)->q.head->next;	\
			if ((_q)->q.head == NULL)		\
				(_q)->q.tail = NULL;		\
			else					\
				(_q)->q.head->prev = NULL;	\
		}						\
		_lock_release(&(_q)->q.lock);			\
	}							\
}


#define _dequeue_end(_q, _thread) {				\
	_lock_acquire(&(_q)->q.lock);				\
	(*(_thread)) = (_q)->q.tail;				\
	if ((_q)->q.tail != NULL) {				\
		(_q)->q.tail = (_q)->q.tail->prev;		\
		if ((_q)->q.tail == NULL)			\
			(_q)->q.head = NULL;			\
		else						\
			(_q)->q.tail->next = NULL;		\
	}							\
	_lock_release(&(_q)->q.lock);				\
}

#endif
