/* This file is part of the LWRPC library */
#include <lwrpc_internal.h>
#include <lwrpc.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

extern void scheduler_loop ();
int __created = 0;
extern MPI_Comm comm_out;
void *_lwrpc_worker (void *arg)
{
	int vp_id = (int) arg;
	lwrpc_t *rpc = (lwrpc_t *)calloc (1, sizeof(lwrpc_t));
	int test0, test1, test2;

	rpc->vp_id = vp_id;
	_lock_init (&rpc->lock);
	rpc->work = scheduler_loop;
	rpc->c_routine = 1;

	scheduler_rpc[rpc->vp_id] = rpc;
	cleanup_rpc[rpc->vp_id][0] = NULL;
	__created++;

	worker_thread[rpc->vp_id] = pthread_self();

	test0 = (entry_routine == NULL);
	test1 = (execution_mode == MODE_SPMD)&&(vp_id == 0);
	test2 = (execution_mode == MODE_MS)&&(lwrpc_node_id() == 0)&&(vp_id == 0);

	if (test0 && (test1 || test2)) {
		lwrpc_t *mainrpc = (lwrpc_t *)calloc (1, sizeof(lwrpc_t));

		mainrpc->vp_id = vp_id;
		_lock_init (&mainrpc->lock);

		uth_vp_init(rpc->vp_id, mainrpc);
		mainrpc->uth_handler = uth_self_ex(mainrpc->vp_id);

		while (__created < kthreads) sched_yield();

		MPI_Barrier(MPI_COMM_WORLD);

		return 0;	/* continue in main */
	}
	else {
		uth_vp_init(rpc->vp_id, rpc);
		rpc->uth_handler = uth_self_ex(rpc->vp_id);

		while (__created < kthreads) sched_yield();

		if (vp_id == 0) MPI_Barrier(MPI_COMM_WORLD);

		scheduler_loop();	/* never returns */
	}

	return 0;
}

void start_worker(int id)
{
	int res;
	pthread_t pth;
	pthread_attr_t attr;
	pthread_attr_init(&attr);

	pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
	res = pthread_create(&pth, &attr, _lwrpc_worker, (void *)id);
	if (res == 0)
		worker_thread[id] = pth;
	else
		Error("pthread_create failed");
}

void shutdown_worker(int id)
{
	int this_node = lwrpc_node_id();

	node_info[this_node].nworkers--;
}

void _lwrpc_md_init()
{
	unsigned int i;
	int this_node = lwrpc_node_id();

	if (lwrpc_num_nodes() > 1) {
		start_server_thread();
	}

	for (i = 1; i<kthreads; i++) {
		start_worker(i);
	}
}


void _lwrpc_md_end ()
{
	unsigned int i;
	unsigned int my_vp;
	int res;

	my_vp = uth_get_vpid();

	if(my_vp!=0) {
		pthread_exit(0);
	}

	if (my_vp==0) {
		for (i=1; i<kthreads; i++) {
			res = pthread_join (worker_thread[i], NULL);
		}

		enter_comm_cs();
		MPI_Barrier(comm_out);
		leave_comm_cs();

		if (lwrpc_num_nodes() > 1) {
			shutdown_server_thread();
		}

		_lwrpc_stats();

		MPI_Barrier(comm_out);
		MPI_Finalize();
		exit(0);
	}
}


void thread_sleep(int ms)
{
	struct timespec req, rem;

	req.tv_sec = ms / 1000;
	req.tv_nsec = (ms % 1000)*1E6;
	nanosleep(&req, &rem);
}

static pthread_mutex_t waitlock = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t waitcond = PTHREAD_COND_INITIALIZER;

void _thread_wait()
{
	pthread_mutex_lock(&waitlock);
	pthread_cond_wait(&waitcond, &waitlock);
	pthread_mutex_unlock(&waitlock);
}

void thread_wait()
{
	if (block_idle_threads) {
		_thread_wait();
	}
}

void _thread_signal()
{
	pthread_mutex_lock(&waitlock);
	pthread_cond_broadcast(&waitcond);
	pthread_mutex_unlock(&waitlock);
}

void thread_signal()
{
	if (block_idle_threads) {
		_thread_signal();
	}
}
