/* This file is part of the LWRPC library */
#ifndef _UTH_TO_PTH_INCLUDED
#define _UTH_TO_PTH_INCLUDED
#include <pthread.h>

struct _uth_t {
	void 		(*func)(void *arg);
	void 		*arg;
	pthread_t	pth_handler;
	pthread_cond_t	pth_cond;
	pthread_mutex_t	pth_mutex;
	int		vp_id;
	int		stacksize;
	struct _uth_t	*prev;
	struct _uth_t	*next;
} ;

typedef struct _uth_t * uth_t;

#ifdef __cplusplus
extern "C" {
#endif

/* Primary API */

extern int uth_init(int stacksize);
extern int uth_vp_init(int vp, void *arg);
extern int uth_get_vpid(void);
extern uth_t uth_create(void (*start)(void *arg), void *arg);
void uth_reinit(uth_t thread, void (*start)(void *arg), void *arg);
extern void uth_delete(uth_t t);

extern uth_t uth_self(void);
extern void *uth_getarg(uth_t thread);

extern uth_t uth_self_ex(int vp);

extern double uth_gettime(void);



/* Extended API */
extern void uth_switchto_ex(int vp, uth_t old, uth_t new);


#ifdef __cplusplus
}           
#endif

#endif
